
class Format:
    def __init__(self, regex):
        self.regex = regex
        self.sims = {'(': 1, '|': 2, '.': 3, '*': 4, '+': 4, '?': 4}


    def prec(self, value):
        return 5 if value.isalnum() else self.sims[value]


    def idempotenciesApp(self):
        regexStr = self.regex
        i = 0
        while i < len(regexStr)-1:
            if regexStr[i] == regexStr[i+1] and regexStr[i] in "+?":
                regexStr = regexStr[:i] + regexStr[i+1:]
            else:
                i += 1
        self.regex = regexStr


    def positiveId(self, regexx):
        expression = regexx
        count_plus = expression.count('+')
        count_plus_lit = expression.count("'+'")
        count_t = count_plus - count_plus_lit
        while count_t > 0:

            count_plus = expression.count('+')
            count_plus_lit = expression.count("'+'")
            count_t = count_plus - count_plus_lit
            
            for i in range(len(expression)):
                if expression[i] == '+':
                    if expression[i-1] == ')':
                        j = i-2
                        continuee = True
                        closeParen = 1
                        while continuee:
                            if expression[j] == ')':
                                closeParen += 1
                            elif expression[j] == '(':
                                closeParen -= 1
                            j -= 1
                            if closeParen == 0:
                                continuee = False
                        expression = f'{expression[:j+1]}{expression[j+1:i]*2}*{expression[i+1:]}'

                    elif expression[i-1].isalnum():
                        before = expression[:i-1]
                        after = expression[i+1:]
                        middle = expression[i-1]*2
                        expression = f'{before}{middle}*{after}'    
        return expression

    
    def zeroOrOneId(self, regexx):
        expression = regexx
        count_q = expression.count('?')
        count_q_lit = expression.count("'?'")
        count_t = count_q - count_q_lit
        while count_t > 0:

            count_q = expression.count('?')
            count_q_lit = expression.count("'?'")
            count_t = count_q - count_q_lit

            for i in range(len(expression)):

                

                if expression[i] == '?':
                    if expression[i-1] == ')':
                        j = i-2
                        continuee = True
                        closeParen = 1
                        while continuee:
                            if expression[j] == ')':
                                closeParen += 1
                            elif expression[j] == '(':
                                closeParen -= 1
                            j -= 1
                            if closeParen == 0:
                                continuee = False
                        expression = f'{expression[:j+1]}({expression[j+1:i]}|ε){expression[i+1:]}'

                    elif expression[i-1].isalnum():
                        before = expression[:i-1]
                        after = expression[i+1:]
                        middle = expression[i-1]
                        expression = f'{before}({middle}|ε){after}' 
                    elif i-2 >= 0:
                        ck = expression[i-3:i]
                        if ck[0] == "'" and ck[-1] == "'":
                            expression = f'{expression[:i-3]}({ck}|ε){expression[i+1:]}'

                               

                    
        return expression


    def concat(self, regexx):
        newRegex, ops = "", list(self.sims.keys())
        ops.remove('(')

        i = 0
        while i < len(regexx):
            val = regexx[i]
            if i + 2 < len(regexx):
                if regexx[i] == "'" and regexx[i + 2] == "'":
                    val = regexx[i + 1]
                    val = str(ord(val))
                    if len(val) == 2:
                        val = '0' + val
                    elif len(val) == 1:
                        val = '00' + val
                    i += 2
                elif val.isalnum() or val in ["#", '_', '^', '"']:
                    val = str(ord(val))
                    if len(val) == 2:
                        val = '0' + val
                    elif len(val) == 1:
                        val = '00' + val

            elif regexx[i].isalnum() or regexx[i] in ['#', '_', '^', '"']:
                val = str(ord(val)) 
                if len(val) == 2:
                    val = '0' + val
                elif len(val) == 1:
                    val = '00' + val
                    
            if i + 1 < len(regexx):
                val_p1 = regexx[i + 1]
                newRegex += val

                if val != '(' and val_p1 != ')' and val != '|' and val_p1 not in ops:
                    newRegex += '.'

            i += 1

        if regexx[-1] == ')' or regexx[-1] in ops:
            newRegex += regexx[-1]
        else:
            if len(str(ord(regexx[-1]))) == 1:
                newRegex += '00' + str(ord(regexx[-1]))
            elif len(str(ord(regexx[-1]))) == 2:
                newRegex += '0' + str(ord(regexx[-1]))

        return newRegex


    def infixPostfix(self):
        postfix, stack = '', []
        concatRegex = self.concat()

        for value in concatRegex:
            if value == '(':
                stack.append(value)

            elif value == ')':
                while stack[-1] != '(':
                    postfix += stack.pop()
                stack.pop()

            else:
                while stack and self.prec(value) <= self.prec(stack[-1]):
                    postfix += stack.pop()
                stack.append(value)

        while stack:
            postfix += stack.pop()
        return postfix



# a = Format("12c++(d|q)??e")
# a.zeroOrOneSus()
# a.positiveSus()
# print(a.regex)
# a.idempotenciesApp()
# a.zeroOrOneId()
# a.positiveId()
# print(a.infixPostfix())
# print(a.regex)




class StateAFD:

    def __init__(self, name, transitions, start=False, accepting=False, value=None):
        self.name = name
        self.transitions = transitions
        self.accepting = accepting
        self.start = start
        self.value = value



from sara_compis1_tools.AFD_tools import AFD_tools
from sara_compis1_tools.StateAFD import StateAFD
from sara_compis1_tools.Error import Error


class LexEval:
    def __init__(self, filename):
        self.filename = filename
        self.file = open(self.filename, 'r', encoding='utf-8')
        self.lines = self.file.readlines()
        self.file.close()

 
    def evaluate(self, mega, prev_errors=None):
        lines =  [line[:-1] if line[-1] == '\n' else line for line in self.lines]
        afd_tools = AFD_tools()
        tokens = []
        errors = prev_errors if prev_errors else set() 
        afd_tools = AFD_tools()
        symbols = afd_tools.get_symbols(mega)

        for line_no, line in enumerate(lines, start=1):
            i = 0
            lenn = len(line)
            while i < lenn:
                if line[i] not in [' ', '\t', '\n']:
                    start = i
                    if line[i] == '"': 
                        i += 1
                        while i < lenn and line[i] != '"':  
                            current = line[i]
                            if current not in symbols:
                                errors.add(Error(line_no, f'Caracter no reconocido: {current}', i))

                            i += 1
                        if i < lenn and line[i] == '"':  
                            i += 1
                    else:
                        while i < lenn and line[i] not in [' ', '\t', '\n']:
                            current = line[i]
                            if current not in symbols:
                                errors.add(Error(line_no, f'Caracter no reconocido: {current}', i))

                            i += 1
                    lexeme = line[start:i]
                else:
                    lexeme = line[i]
                    i += 1

                accepted = afd_tools.afn_simulation(mega, lexeme)
                if accepted:
                    if accepted != 'empty_value':
                        tokens.append(accepted)
                else:
                    if lexeme not in [' ', '\t', '\n']:
                        errors.add(Error(line_no, f'Lexema no aceptado: {lexeme}', start))
        return tokens, errors
    

    def sort_errors(self, errors):
        errors = list(errors)
        errors.sort(key=lambda x: x.line)
        return errors


    def print_tokens(self, result_sim):
        tokens, errors = result_sim
        if not errors:
            print('')
            for token in tokens:
                exec(token)
        else:
            print('\n')
            errors = self.sort_errors(errors)
            for error in errors:
                if error.position:
                    print(f'Error en línea {error.line}: \n{error.error}, posición {error.position}\n')
                else:   
                    print(f'Error en línea {error.line}: \n{error.error}\n')





import networkx as nx
from graphviz import Digraph

class Visualizer:
    def __init__(self):
        pass

    def draw_mega_afd(self, afd):
        G = nx.MultiDiGraph()
        for state in afd:
            node_attrs = {'shape': 'circle'}
            if state.start:
                node_attrs.update({'color': 'green', 'style': 'filled'})
            if state.accepting:
                node_attrs.update({'peripheries': '2'})
            G.add_node(str(state.name), **node_attrs)

            for transition, final_dest in state.transitions.items():
                G.add_node(str(final_dest))
                
                if ord(transition) in [el for el in range(0, 35)]:
                    transition = str(ord(transition))

                    if len(transition) == 1:
                        transition = '00' + transition
                    elif len(transition) == 2:
                        transition = '0' + transition

                

                G.add_edge(str(state.name), str(final_dest), label=transition, dir='forward')    

        dot = Digraph()
        for u, v, data in G.edges(data=True):
            dot.edge(u, v, label=data['label'], dir=data['dir'])
        for node in G.nodes:
            attrs = G.nodes[node]
            dot.node(node, **attrs)
        dot.render('mega/megaautomata', format='png')




class Syntax:
    def __init__(self, string): 
        self.string = string
        self.sims = {'(': 1, '|': 2, '.': 3, '*': 4, '+': 4, '?': 4}
        

    def checkParenthesis(self):
        stack = []
        for i in range(len(self.string)):
            if self.string[i] == '(': 
                stack.append(self.string[i])
            elif self.string[i] == ')':
                if not stack:
                    return False
                stack.pop()
        return not stack


    def checkOperator(self):
        for i in range(len(self.string)):
            if self.string[i] in self.sims.keys():
                if self.string[i] != '(':
                    if i == 0:
                        return False
        return True


    def checkOperatorValid(self):
        for i in range(len(self.string)):
            if not self.string[i].isalnum() and self.string[i] not in self.sims.keys() and self.string[i] != ')':
                return False
        return True
    
    
    def checkDot(self):
        for i in range(len(self.string)):
            if(self.string[i] == '.'):
                return False
        return True
    

    def checkMultU(self):
        count = 0
        for c in self.string:
            if c == '|':
                count += 1
                if count >= 2:
                    return True
            else:
                count = 0

    def checkLastNotU(self):
        if self.string[-1] == '|':
            return False
        return True
    


import os
import networkx as nx
from graphviz import Digraph

from Format import Format
os.environ["PATH"] += os.pathsep + 'C:/Program Files (x86)/Graphviz/bin'    
from StateAFD import StateAFD


class Node:
    def __init__(self, symbol, parent = None, left = None, right = None, no = None, anulable = False, firstpos = [], lastpos = []):
        self.symbol = symbol
        self.parent = parent
        self.left = left
        self.right = right
        self.no = no  
        self.anulable = anulable
        self.firstpos = firstpos
        self.lastpos = lastpos

class npObj:
    def __init__(self, treeNo, symbol, nextpos = []):
        self.symbol = symbol
        self.nextpos = nextpos        
        self.treeNo = treeNo


class AFD:
    def __init__(self, regex):
        self.regex = regex
        self.tree = None
        self.table = {}
        self.tableSet = set()
        self.transitions = {}

    def augmentRegex(self):
        hashRegex = Format(self.regex + '#')
        # hashRegex.idempotenciesApp()
        a = hashRegex.positiveId(self.regex)
        b = hashRegex.zeroOrOneId(a)
        return hashRegex.concat(b)
    
    def syntaxTree(self):
        tree = []
        toDo = []
        enum = 1
        regex = self.regex.regex
        subexpr_stack = [] 

        i = 0
        while i < len(regex):
            rrrr = regex[i]
            if len(toDo) > 0:
                if toDo[-1] == '|':
                    if len(tree) > 1:
                        l = tree.pop(0)
                        r = tree.pop(0)
                        newSymU = Node(toDo.pop(), left=l, right=r)
                        l.parent = newSymU
                        r.parent = newSymU
                        tree.append(newSymU)
                elif toDo[-1] == '.':
                    if len(tree) > 1:
                        l = tree.pop(0)
                        r = tree.pop(0)
                        newSymC = Node(toDo.pop(), left=l, right=r)
                        l.parent = newSymC
                        r.parent = newSymC
                        tree.append(newSymC)
            if regex[i].isalnum() or regex[i:i+3] == '035':
                if i+2+1 < len(regex) and regex[i+2+1] == '*':
                    if regex[i:i+3] != '949':
                        alnumNode = Node(regex[i:i+3], no=enum)
                        kleeneNode = Node(regex[i+2+1], left=alnumNode)
                        alnumNode.parent = kleeneNode
                        tree.append(kleeneNode)
                        enum += 1
                        i += 2
                    else:
                        alnumNode = Node(regex[i:i+3])
                        kleeneNode = Node(regex[i+2+1], left=alnumNode)
                        alnumNode.parent = kleeneNode
                        tree.append(kleeneNode)
                        i += 2
                else:
                    ehh = regex[i:i+3]
                    if regex[i:i+3] != '949':
                        alnumNode = Node(regex[i:i+3], no=enum)
                        tree.append(alnumNode)
                        enum += 1
                        i += 2
                    else:
                        alnumNode = Node(regex[i:i+3])
                        tree.append(alnumNode)
                        i += 2
            elif regex[i] == '(':
                subexpr_stack.append(tree)  
                tree = []  
            elif regex[i] == ')':
                if len(subexpr_stack) > 0:
                    parent_tree = subexpr_stack.pop()  
                    if len(tree) > 0:
                        if regex[i+1] == '*':
                            parent_tree.append(Node(regex[i+1], left=tree[0]))
                            i += 1
                        else:
                            parent_tree.append(tree[0])  
                    tree = parent_tree  
            elif regex[i] == '|' or regex[i] == '.':
                if len(tree) < 2:
                    toDo.append(regex[i])
            i += 1

        while toDo and tree:
            if toDo[-1] == '|':
                if len(tree) > 1:
                    l = tree.pop(0)
                    r = tree.pop(0)
                    newSymU = Node(toDo.pop(), left=l, right=r)
                    l.parent = newSymU
                    r.parent = newSymU
                    tree.append(newSymU)
            elif toDo[-1] == '.':
                if len(tree) > 1:
                    l = tree.pop(0)
                    r = tree.pop(0)
                    newSymC = Node(toDo.pop(), left=l, right=r)
                    l.parent = newSymC
                    r.parent = newSymC
                    tree.append(newSymC)
        return tree
    

    def anulable(self, tree):
        if tree:
            self.anulable(tree.left)
            self.anulable(tree.right)
            if tree.symbol == '949':
                tree.anulable = True
            elif tree.symbol.isalnum():
                tree.anulable = False
            elif tree.symbol == '|':
                    tree.anulable = tree.left.anulable or tree.right.anulable
            elif tree.symbol == '.':
                    tree.anulable = tree.left.anulable and tree.right.anulable
            elif tree.symbol == '*':
                tree.anulable = True
        return tree


    def firstPosMethod(self, tree):
        if tree:
            self.firstPosMethod(tree.left)
            self.firstPosMethod(tree.right)
            if tree.symbol.isalnum() and tree.no or tree.no == 0 or tree.symbol == '035':
                tree.firstpos = [tree.no]
            if tree.symbol == '|':
                tree.firstpos = tree.left.firstpos + tree.right.firstpos
            if tree.symbol == '.':
                if tree.left.anulable:
                    tree.firstpos = tree.left.firstpos + tree.right.firstpos
                else:
                    tree.firstpos = tree.left.firstpos
            if tree.symbol == '*':
                tree.firstpos = tree.left.firstpos
        return tree


    def lastPosMethod(self, tree):
        if tree:
            self.lastPosMethod(tree.left)
            self.lastPosMethod(tree.right)
            if tree.symbol.isalnum() and tree.no or tree.no == 0 or tree.symbol == '035':
                tree.lastpos = [tree.no]
            if tree.symbol == '|':
                tree.lastpos = tree.left.lastpos + tree.right.lastpos
            if tree.symbol == '.':
                if tree.right.anulable:
                    tree.lastpos = tree.left.lastpos + tree.right.lastpos
                else:
                    tree.lastpos = tree.right.lastpos
            if tree.symbol == '*':
                tree.lastpos = tree.left.lastpos
        return tree
    

    def genNextPosDict(self, tree):
        if tree:
            self.genNextPosDict(tree.left)
            self.genNextPosDict(tree.right)
            if tree.no or tree.no == 0:
                self.table[tree.no] = {tree.symbol: []}


    def genNextPos(self, tree):
        if tree:
            self.genNextPos(tree.left)
            self.genNextPos(tree.right)
            if tree.symbol == '.':
                for i in tree.left.lastpos:
                    for key in self.table[i]:
                        if tree.right.firstpos not in self.table[i][key]:
                            self.table[i][key] += tree.right.firstpos
            if tree.symbol == '*':
                for i in tree.lastpos:
                    for key in self.table[i]:
                        if tree.firstpos not in self.table[i][key]:
                            self.table[i][key] += tree.firstpos
        

    def tableToObj(self):
        for key in self.table:
            for key2 in self.table[key]:
                s = key2
                nP = self.table[key][key2]
                self.tableSet.add(npObj(treeNo=key, symbol=s, nextpos=nP))
        return self.tableSet
        

    def genAFD(self, count_states = 0):
        table = self.tableSet
        states = [self.tree.firstpos]
        toDo = [self.tree.firstpos]
        newAFD = []
        acceptState = None

        count = 0
        while toDo:
            toDoState = toDo.pop(0)
            symbols = {} 
            for elem in toDoState:
                for elem2 in table:
                    if elem == elem2.treeNo:
                        if elem2.symbol != '035':
                            if elem2.symbol not in symbols:
                                symbols[elem2.symbol] = set(elem2.nextpos)
                            else:
                                symbols[elem2.symbol].update(elem2.nextpos)
                            if list(symbols[elem2.symbol]) not in states:
                                states.append(list(symbols[elem2.symbol]))
                                toDo.append(list(symbols[elem2.symbol]))
                        else:
                            acceptState = elem2.treeNo
            newState = StateAFD(name=toDoState, transitions=symbols)
            if not count:
                newState.start = True
                count += 1
            newAFD.append(newState)

        for elem in newAFD:
            if acceptState in elem.name:
                elem.accepting = True
        
        # Changing the name of the states
        count = count_states
        for state in newAFD:
            for st in newAFD:
                for key, transition in st.transitions.items():
                    if transition == set(state.name):
                        st.transitions[key] = chr(65+count)
            
            state.name = chr(65+count)
            count += 1

        return newAFD
    

    def minimizationAFD(self):
        # Creando copia del AFD
        afd = self.genAFD()

        # Unir estados de aceptacion y que no son de aceptacion
        accepting_states = set(state for state in afd if state.accepting)
        non_accepting_states = set(state for state in afd if not state.accepting)
        state_groups = [accepting_states, non_accepting_states]

        # Repetir hasta que no se puedan unir mas estados
        while True:
            new_state_groups = []
            for group in state_groups:
                # Por cada grupo de estados, agrupar por transiciones
                transition_groups = {}
                for state in group:
                    transition = tuple(sorted(state.transitions.values()))
                    if transition not in transition_groups:
                        transition_groups[transition] = set()
                    transition_groups[transition].add(state)

                # Por cada grupo de transiciones, unir estados
                for transition_group in transition_groups.values():
                    if len(transition_group) > 1:
                        new_state_groups.append(transition_group)
                    else:
                        new_state_groups.append({transition_group.pop()})

            # Si ya no se pueden unir mas estados, terminar
            if len(new_state_groups) == len(state_groups):
                break
            state_groups = new_state_groups

        # Crear nuevo AFD
        statesI = sum(len(group) for group in state_groups)
        reps = {}
        for group in state_groups:
            if len(group) > 1:
                same = []
                for element in group:
                    same.append(element)
                reps[chr(65+statesI)] = tuple(same)
                statesI += 1


        for replacement, same in reps.items():
            check = tuple([obj.name for obj in same])
            checkAccepting = tuple([obj.accepting for obj in same])
            checkStart = tuple([obj.start for obj in same])
            for key, state in afd.items():
                for k, v in state.transitions.items():
                    if v in check:
                        state.transitions[k] = replacement
                if  checkAccepting.count(True) > 0:
                    if state.name in check:
                        state.accepting = True
                if checkStart.count(True) > 0:
                    if state.name in check:
                        state.start = True
                if state.name in check:
                    state.name = replacement

        miniAFD = {}
        index = 0
        for state in afd:
            if state.name not in [obj.name for obj in miniAFD.values()]:
                miniAFD[index] = state
                index += 1

        return miniAFD



    def draw_afd(self, afd):

        G = nx.MultiDiGraph()
        for state in afd:
            node_attrs = {'shape': 'circle'}
            if state.start:
                node_attrs.update({'color': 'green', 'style': 'filled'})
            if state.accepting:
                node_attrs.update({'peripheries': '2'})
            G.add_node(str(state.name), **node_attrs)

            for transition, final_dest in state.transitions.items():
                G.add_node(str(final_dest))
                G.add_edge(str(state.name), str(final_dest), label=str(chr(int(transition))), dir='forward')

        dot = Digraph()
        for u, v, data in G.edges(data=True):
            dot.edge(u, v, label=data['label'], dir=data['dir'])
        for node in G.nodes:
            attrs = G.nodes[node]
            dot.node(node, **attrs)

        dot.attr(rankdir='LR')
        dot.render('directAFD/directAFD', format='png')


    def draw_mini_afd(self):
        afd = self.minimizationAFD()

        G = nx.MultiGraph()
        for state in afd.values():
            if state.start:
                G.add_node(state.name, color='green', style='filled', shape='circle')
            if state.accepting:
                G.add_node(state.name, shape='doublecircle')
            for k, v in state.transitions.items():
                if state.start:
                    G.add_node(state.name, color='green', style='filled', shape='circle')
                if state.accepting:
                    G.add_node(state.name, shape='doublecircle')
                else:
                    if v != 'estado muerto':
                        G.add_node(v)
                if v != 'estado muerto':
                    G.add_edge(state.name, v, label=k, dir='forward')
                   
        dot = Digraph()
        for u, v, data in G.edges(data=True):
            dot.edge(u, v, label=data['label'], dir=data['dir'])
        for node in G.nodes:
            attrs = G.nodes[node]
            dot.node(node, **attrs)

        dot.attr(rankdir='LR')
        dot.render('directAFD/miniDirectAFD', format='png')


    
    def simulateDirectAFD(self, string, afd):
        current_state = afd[0]
        for symbol in string:
            if symbol not in current_state.transitions:
                return False
            current_state = [afd[i] for i in range(len(afd)) if afd[i].name == current_state.transitions[symbol]]
            if not current_state:
                return False
            else:
                current_state = current_state[0]
        return current_state.accepting
    

    def simulateMiniAFD(self, string):
        afd = self.minimizationAFD()
        current_state = afd[0]
        for symbol in string:
            if symbol not in current_state.transitions:
                return False
            current_state = [state for key, state in afd.items() if state.name == current_state.transitions[symbol]]
            if not current_state:
                return False
            else:
                current_state = current_state[0]
        return current_state.accepting


    
    def generateAFD(self, count_states):
        st = self.syntaxTree()
        anulable = self.anulable(st[0])
        fP = self.firstPosMethod(anulable)
        lP = self.lastPosMethod(fP)
        self.tree = lP
        treeVar = self.tree
        self.genNextPosDict(treeVar)
        self.genNextPos(treeVar)
        self.tableToObj()
        return self.genAFD(count_states)
        # self.draw_afd(data)

    
    def generateMiniAFD(self):
        st = self.syntaxTree()
        anulable = self.anulable(st[0])
        fP = self.firstPosMethod(anulable)
        lP = self.lastPosMethod(fP)
        self.tree = lP
        treeVar = self.tree
        self.genNextPosDict(treeVar)
        self.genNextPos(treeVar)
        self.tableToObj()
        self.draw_mini_afd()


    def simulateDirectAFD_General(self, miniString):
        st = self.syntaxTree()
        anulable = self.anulable(st[0])
        fP = self.firstPosMethod(anulable)
        lP = self.lastPosMethod(fP)
        self.tree = lP
        treeVar = self.tree
        self.genNextPosDict(treeVar)
        self.genNextPos(treeVar)
        self.tableToObj()
        data = self.genAFD()
        if self.simulateDirectAFD(miniString, data):
            print('Simulacion AFD Directo: Cadena aceptada')
        else:
            print('Simulacion AFD Directo: Cadena no aceptada')


    def simulateMiniAFD_General(self, miniString):
        st = self.syntaxTree()
        anulable = self.anulable(st[0])
        fP = self.firstPosMethod(anulable)
        lP = self.lastPosMethod(fP)
        self.tree = lP
        treeVar = self.tree
        self.genNextPosDict(treeVar)
        self.genNextPos(treeVar)
        self.tableToObj()
        data = self.genAFD()
        if self.simulateMiniAFD(miniString):
            print('Simulacion AFD Minimizado: Cadena aceptada')
        else:
            print('Simulacion AFD Minimizado: Cadena no aceptada')


        
        
    

    def generatelP(self):
        st = self.syntaxTree()
        anulable = self.anulable(st[0])
        fP = self.firstPosMethod(anulable)
        lP = self.lastPosMethod(fP)
        return lP
    

    def defineInitialAndAceptting(self, table, initial, aceptting):
        for k, v in table.items():
            if v.positions == initial:
                v.initial = True
            if v.positions in aceptting:
                v.aceptting = True
        return table





    def printVisualTree(self, tree, level=0):
        if tree:
            self.printVisualTree(tree.right, level+1)
            if tree.no or tree.no == 0:
                print('  '*(level*3) + str(tree.symbol))
            else:
                print('  '*(level*3) + str((tree.symbol)))
            self.printVisualTree(tree.left, level+1)


def printPostOrder(tree):
    if tree:
        printPostOrder(tree.left)
        printPostOrder(tree.right)
        print(tree.symbol)



# string = '(0|1|2)#'
# afdd = AFD(string)
# tree = afdd.syntaxTree()[0]
# afdd.generateAFD()


from StateAFD import StateAFD

class AFD_tools:
    def __init__(self):
        pass


    get_state = lambda self, name, states: next(filter(lambda state: state.name == name, states), None)

    def epsilon_clousure(self, state, states):
        
        clousure = {state}
        if 'ε' in state.transitions:
            for target in state.transitions['ε']:
                next_state = self.get_state(target, states)
                clousure |= self.epsilon_clousure(next_state, states)

        return clousure
 

    def move(self, states_set, sym, states):
        res = set()

        for state in states_set:
            if sym in state.transitions:
                for target in state.transitions[sym]:
                    next_state = self.get_state(target, states)
                    res.add(next_state)
        return res


    def afn_simulation(self, afn, str_eval):
        current_states = set(filter(lambda s: s.start, afn))
        epsilon_states = set()

        for state in current_states:
            epsilon_states |= self.epsilon_clousure(state, afn)

        for sym in str_eval:
            next_states = set()

            for state in epsilon_states:
                next_states |= self.move({state}, sym, afn)

            epsilon_states = set()
            for state in next_states:
                epsilon_states |= self.epsilon_clousure(state, afn)

        for state in epsilon_states:
            if state.accepting:
                return state.value if state.value else 'empty_value'


    def get_symbols(self, afn):
        symbols = set()
        for state in afn:
            for sym in state.transitions:
                if sym != 'ε':
                    symbols.add(sym)
        return list(symbols)
    

if __name__ == '__main__':
    afd_tools = AFD_tools()
    mega = [StateAFD(name='init',transitions={'ε': 'A'},accepting=False,start=True, value=None),StateAFD(name='A',transitions={'a': 'B', 'b': 'B', 'c': 'B', 'A': 'B', 'B': 'B', 'C': 'B'},accepting=False,start=False, value=None),StateAFD(name='B',transitions={'a': 'B', 'b': 'B', 'c': 'B', 'A': 'B', 'B': 'B', 'C': 'B', '0': 'B', '1': 'B', '2': 'B', '3': 'B', '4': 'B', '5': 'B', '6': 'B', '7': 'B', '8': 'B', '9': 'B'},accepting=True,start=False, value='print("Identificador\\n")'),StateAFD(name='init',transitions={'ε': 'C'},accepting=False,start=True, value=None),StateAFD(name='C',transitions={'i': 'D'},accepting=False,start=False, value=None),StateAFD(name='D',transitions={'f': 'E'},accepting=False,start=False, value=None),StateAFD(name='E',transitions={},accepting=True,start=False, value='print("If\\n")'),StateAFD(name='init',transitions={'ε': 'F'},accepting=False,start=True, value=None),StateAFD(name='F',transitions={'f': 'G'},accepting=False,start=False, value=None),StateAFD(name='G',transitions={'o': 'H'},accepting=False,start=False, value=None),StateAFD(name='H',transitions={'r': 'I'},accepting=False,start=False, value=None),StateAFD(name='I',transitions={},accepting=True,start=False, value='print("For\\n")'),StateAFD(name='init',transitions={'ε': 'J'},accepting=False,start=True, value=None),StateAFD(name='J',transitions={'w': 'K'},accepting=False,start=False, value=None),StateAFD(name='K',transitions={'h': 'L'},accepting=False,start=False, value=None),StateAFD(name='L',transitions={'i': 'M'},accepting=False,start=False, value=None),StateAFD(name='M',transitions={'l': 'N'},accepting=False,start=False, value=None),StateAFD(name='N',transitions={'e': 'O'},accepting=False,start=False, value=None),StateAFD(name='O',transitions={},accepting=True,start=False, value='print("While\\n")'),StateAFD(name='init',transitions={'ε': 'P'},accepting=False,start=True, value=None),StateAFD(name='P',transitions={'0': 'Q', '1': 'Q', '2': 'Q', '3': 'Q', '4': 'Q', '5': 'Q', '6': 'Q', '7': 'Q', '8': 'Q', '9': 'Q'},accepting=False,start=False, value=None),StateAFD(name='Q',transitions={'0': 'Q', '1': 'Q', '2': 'Q', '3': 'Q', '4': 'Q', '5': 'Q', '6': 'Q', '7': 'Q', '8': 'Q', '9': 'Q'},accepting=True,start=False, value='print("Entero\\n")'),StateAFD(name='init',transitions={'ε': 'R'},accepting=False,start=True, value=None),StateAFD(name='R',transitions={'+': 'S', '-': 'S', '*': 'S', '/': 'S'},accepting=False,start=False, value=None),StateAFD(name='S',transitions={},accepting=True,start=False, value='print("Operador aritmetico\\n")')]
    print(afd_tools.afn_simulation(mega, 'if'))


from sara_compis1_tools.StateAFD import StateAFD
from lexEval import LexEval
from Error import Error
import sys

mega = [StateAFD(name='init',transitions={'ε': 'A'},accepting=False,start=True, value=None),StateAFD(name='A',transitions={'0': 'B', '1': 'B', '2': 'B', '3': 'B', '4': 'B', '5': 'B', '6': 'B', '7': 'B', '8': 'B', '9': 'B'},accepting=False,start=False, value=None),StateAFD(name='B',transitions={'0': 'B', '1': 'B', '2': 'B', '3': 'B', '4': 'B', '5': 'B', '6': 'B', '7': 'B', '8': 'B', '9': 'B'},accepting=True,start=False, value='print("Número entero\\n")'),StateAFD(name='init',transitions={'ε': 'C'},accepting=False,start=True, value=None),StateAFD(name='C',transitions={'0': 'D', '1': 'D', '2': 'D', '3': 'D', '4': 'D', '5': 'D', '6': 'D', '7': 'D', '8': 'D', '9': 'D'},accepting=False,start=False, value=None),StateAFD(name='D',transitions={'0': 'D', '1': 'D', '2': 'D', '3': 'D', '4': 'D', '5': 'D', '6': 'D', '7': 'D', '8': 'D', '9': 'D', '.': 'E'},accepting=False,start=False, value=None),StateAFD(name='E',transitions={'0': 'F', '1': 'F', '2': 'F', '3': 'F', '4': 'F', '5': 'F', '6': 'F', '7': 'F', '8': 'F', '9': 'F'},accepting=False,start=False, value=None),StateAFD(name='F',transitions={'0': 'F', '1': 'F', '2': 'F', '3': 'F', '4': 'F', '5': 'F', '6': 'F', '7': 'F', '8': 'F', '9': 'F'},accepting=True,start=False, value='print("Número decimal\\n")')]
errors = set([Error(line=11, error='Error: Token no definido: identificador'),Error(line=16, error='Error: Token no definido: cadena'),Error(line=13, error='Error: Token no definido: palabra_reservada_for'),Error(line=12, error='Error: Token no definido: palabra_reservada_if')])

if len(sys.argv) < 2:
	print('Por favor ingrese el archivo plano')
	sys.exit(1)
txt_file = sys.argv[1]

lex = LexEval(txt_file)
results = lex.evaluate(mega, errors)
lex.print_tokens(results)

from Visualizer import Visualizer
v = Visualizer()
v.draw_mega_afd(mega)



from setuptools import setup, find_packages

setup(
    name='sara_compis1_tools',
    version='0.0.10',
    description='A collection of tools for the Language Design course',
    long_description=open('README.txt').read() + '\n\n' + open('CHANGELOG.txt').read(),
    url='https://github.com/MGonza20/Compis_Lab4',
    author='Sara Paguaga',
    author_email='sara.paguaga@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    keywords='Compiler',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'networkx',
        'graphviz'
    ]
)


from Format import Format
from directAFD import AFD
from StateAFD import StateAFD
from Error import Error
import sys


class Token:
        def __init__(self, name):
            self.name = name
            self.regex = None
            self.line_no = None
            self.value = None
    
        def __str__(self):
            return f"Token({self.name}, {self.regex})"


class ruleLine:
    def __init__(self, line_no, line):
        self.line_no = line_no
        self.line = line


class Lexer:
    def __init__(self, filename):
        self.filename = filename
        self.tokens = []
        self.tokenizer = None

    
    def remove_spaces(self, lines):
        wo_spaces = []
        for line in lines:
            new_line = []
            between_q = False
            for char in line:
                if char == "'":
                    between_q = not between_q
                if char != " " or between_q:
                    new_line.append(char)
            wo_spaces.append("".join(new_line))
        return wo_spaces
    

    def clean_comments(self, joined):
        for index, line in enumerate(joined):
            line_wo_comment = ''
            i = 0
            while i < len(line):
                if i < len(line) - 1 and line[i] == '(' and line[i + 1] == '*':
                    while i < len(line) - 1 and (line[i] != '*' or line[i + 1] != ')'):
                        i += 1
                    i += 2
                else:
                    line_wo_comment += line[i]
                    i += 1
            joined[index] = line_wo_comment
        return joined


    def getLines(self, tokens_mode):
        f = open(self.filename, "r", encoding="utf-8")
        lines = f.readlines()
        f.close()
        
        if tokens_mode:
            lines = [line.encode('utf-8').decode('unicode_escape') for line in lines]

        lines_with_n = [n[:-1] for n in lines]
        check_comments = [lll.split(' ') for lll in lines_with_n]  

        joined = [' '.join(line) for line in check_comments]
        # Revision de errores en comentarios
        for line_no, lj in enumerate(joined, start=1):
            comments_stack = []
            for i in range(len(lj) - 1):

                current_next = lj[i:i+2]
                if current_next == '(*':
                    comments_stack.append(current_next)
                elif current_next == '*)':
                    if comments_stack and comments_stack[-1] == '(*':
                        comments_stack.pop()
                    else:
                        raise Exception(f"Error en comentario, linea {line_no}")
            if comments_stack:
                raise Exception(f"Error en comentario, linea {line_no}")

        joined = self.clean_comments(joined)
            
        return joined
    

    def special_split(self, text, delimiter):
        result = []
        start = 0
        open_curly_b = 0

        for index, char in enumerate(text):
            if char == '{':
                open_curly_b += 1
            elif char == '}':
                open_curly_b -= 1

            elif char == delimiter and not open_curly_b:
                result.append(text[start:index])
                start = index + 1

        result.append(text[start:])
        return result
    

    def remove_spaces_list(self, lines):
        for line in lines:
            if not all(element == '' for element in line):
                while '' in line:
                    for element in line:
                        if element == '':
                            line.pop(line.index(element))
        return lines
    

    def assign_values(self):
        splits = [self.special_split(line, ' ') for line in self.getLines(tokens_mode=False)]
        splits = self.remove_spaces_list(splits)
        
        start = 0
        rules_lines = []
        for indx, line in enumerate(splits, start=1):
            if line[0] == 'rule':
                start = indx + 1
            if start:
                rules_lines.append(ruleLine(indx, line))

        
        if len(rules_lines) > 1:

            rules_lines = rules_lines[1:]
            for rule in rules_lines:
                if rule.line[0] == '|':
                    rule.line.pop(0)
                if len(rule.line) == 1:
                    rule.line = rule.line[0]

            rules_dict = {}
            for rule in rules_lines:
                name = ''
                value = ''
                inside = False

                for i in range(len(rule.line)):
                    if rule.line[i] == '{':
                        inside = True
                        continue

                    if rule.line[i] == '}':
                        inside = False
                        continue

                    if not inside:
                        if rule.line[i] not in ['\t', ' ']:
                            name += rule.line[i]
                    else:
                        value += rule.line[i]

                if name:
                    rules_dict[name] = (ruleLine(rule.line_no, value.strip()))
        return rules_dict


    def getTokens(self):
        lines = self.remove_spaces(self.getLines(tokens_mode=True))
        for line_no, line in enumerate(lines, start=1):
            # generando tokens
            if line[:3] == 'let':
                name, regex = line[3:].split('=')
                token = Token(name)
                token.regex = regex
                token.line_no = line_no
                self.tokens.append(token)


    def range_maker(self, start, end, no):
        if len(start) == 3 and len(end) == 3:
            start, end = start[1], end[1]

        if start.isalpha() and end.isalpha():
            if ord(start) > ord(end):
                raise Exception("Error: Rango incorrecto, linea " + str(no))
            elements = [chr(i) for i in range(ord(start), ord(end) + 1)]
            aaa = 123

        elif start.isdigit() and end.isdigit():
            start, end = int(start), int(end)
            if start > end:
                raise Exception("Error: Rango incorrecto, linea " + str(no))
            elements = [str(i) for i in range(start, end + 1)]
        else:
            raise Exception("Formato de regex incorrecto")

        return elements

    
    def change_range_format(self):
        tokens = self.tokens
        for token in tokens:
            new_regex = ''
            i = 0

            if token.regex[0] == "'" and token.regex[-1] == "'" and token.regex.count("'") == 2:
                unquoted_token = token.regex[1:-1]
                token.regex = f'({unquoted_token})'
                continue
            

            while i < len(token.regex):

                p_left = token.regex.count("(")
                p_right = token.regex.count(")")
                if (p_left + p_right) % 2 != 0:
                    raise Exception("Error: Los parentesis no estan balanceados, "+ "linea " + str(token.line_no))
                
                c_left = token.regex.count("[") 
                c_right = token.regex.count("]") 
                if (c_left + c_right )% 2 != 0:
                    raise Exception("Error: Los corchetesno estan balanceados, "+ "linea " + str(token.line_no))

                
                if token.regex[i] == '[':
                    if token.regex.count("'") % 2 != 0:
                        raise Exception("Error en comillas, " + "linea " + str(token.line_no))
                    
                    content = ""
                    j = i + 1
                    while token.regex[j] != ']':
                        content += token.regex[j]
                        j += 1

                    if content.count("''") > 0:
                        content = content.replace("''", "'|'")
                        tokens_list = content.split("|")

                        elements = []
                        for k in range(len(tokens_list)):
                            if tokens_list[k] == "'-'":
                                continue
                            elif '-' in tokens_list[k]:
                                start, end = tokens_list[k].split('-')
                                elements += self.range_maker(start, end, token.line_no)

                        if elements:
                            content = '|'.join(elements)

                    else:
                        if '-' in content:
                            if content.count('-') > 1:
                                raise Exception("Formato de regex incorrecto,"+ "linea " + str(token.line_no))
                            start, end = content.split('-')
                            elements = self.range_maker(start, end, token.line_no)
                            content = '|'.join(elements)
                        elif '-' not in content:
                            content = content[1:-1]
                            elements = [content[i] for i in range(len(content))]
                            content = '|'.join(elements)
                    new_regex += '(' + content + ')'
                    i = j
                else:
                    check = ""
                    j = i
                    while token.regex[j] not in ['+', '*', '?', '(', ')', '[', ']']:
                        check += token.regex[j]
                        j += 1
                    keys = [tk.name for tk in tokens]
                    if check in keys:
                        i = j - 1
                        new_regex += check
                    else:
                        new_regex += token.regex[i]
                i += 1

            count_all = int((new_regex.count('(') + new_regex.count(')')) /2)
            if not count_all or new_regex[-1] in ['+', '*', '?']:
                if  new_regex[0] == "'" and new_regex[-1] == "'":
                    new_regex = new_regex[1:-1]
                new_regex = f'({new_regex})'

            token.regex = new_regex


    def replace_tokens(self):
        for tk in self.tokens:
            for token in self.tokens:
                index = tk.regex.find(token.name)
                while index != -1:
                    right_side = (index + len(token.name) == len(tk.regex) or not ((tk.regex[index + len(token.name)])).isalnum()) 

                    if right_side:
                        tk.regex = tk.regex[:index] + token.regex + tk.regex[index + len(token.name):]
                    index = tk.regex.find(token.name, index + 1)

    
    def surround_dot(self):
        for token in self.tokens:
            if token.regex.count('.') > 0:
                token.regex = token.regex.replace('.', "'.'" )
    

    def remove_double_parentheses(self, token):
        i = 0
        output = ""
        while i < len(token):
            if token[i] == '(':
                content = "("
                count_open = 1
                count_control = 1
                i += 1
                while count_open != 0:
                    if token[i] == '(':
                        count_open += 1
                        count_control += 1
                    elif token[i] == ')':
                        count_open -= 1
                    content += token[i]
                    i += 1
                
                start = content[:count_control]
                end = content[-count_control:]
                opp = '('*count_control
                clp = ')'*count_control
                if start == opp and end == clp and count_control > 1: 
                    output += content[count_control-1:-count_control+1]
                else:
                    output += content
                
            else:
                output += token[i]
                i += 1
        return output

    
    def generate_automatas(self):
        mega_content = []
        count = 0
        return_values = self.assign_values()
        done = []
        errors = set()
        for token in self.tokens:
            if token.name in return_values:
                ff = Format(token.regex)
                token.regex = ff.positiveId(token.regex + '#')
                token.regex = ff.zeroOrOneId(token.regex)
                token.regex = self.remove_double_parentheses(token.regex)
                token.regex = ff.concat(token.regex)
                token.value = return_values[token.name]
                done.append(token.name)
                
                afdd = AFD(token)
                new_afd = afdd.generateAFD(count)
                for elem in new_afd:
                    if elem.accepting:
                        elem.value = token.value

                mega_content.append(new_afd)
                count += len(new_afd)

        for rt, rt_val in return_values.items():
            if rt not in [tk.name for tk in self.tokens]:
                if rt[0] != "'" and rt[-1] != "'":
                    errors.add(Error(line=rt_val.line_no, error="Error: Token no definido: " + rt + " en archivo .yal"))

        for value in return_values:
            if value[0] == "'" and value[-1] == "'" and value not in done:
                lenn = len(value)
                token = Token(name=value[1:-1])
                token.regex = value + '#'
                token.value = return_values[value]
                ff = Format(token.regex)
                token.regex = ff.concat(token.regex)
                done.append(token.name)
                
                afdd = AFD(token)
                new_afd = afdd.generateAFD(count)
                for elem in new_afd:
                    if elem.accepting:
                        elem.value = token.value

                mega_content.append(new_afd)
                count += len(new_afd)
        return mega_content, errors
    

    def unify(self, mega_content):
        stack = []    
        for element in mega_content:
            for state in element:
                if state.start:
                    state.start = False
                    init_state = StateAFD(name='init', start=True, transitions={})
                    init_state.transitions['949'] = state.name
                    stack.append(init_state)
                stack.append(state)
        return stack
    

    def read(self):
        self.assign_values()
        self.getTokens()
        self.change_range_format()
        self.surround_dot()
        self.replace_tokens()
    

    def change_values(self, mega_automata):
        for state in mega_automata:
            updates = {str(chr(int(key))): state.transitions.pop(key) for key in list(state.transitions.keys())}
            state.transitions.update(updates)


    
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("Por favor ingrese el archivo .yal")
        sys.exit(1)

    yal_file = sys.argv[1]
    lexer = Lexer(yal_file)
    
    lexer.read()
    mega_content, errors = lexer.generate_automatas()
    mega_automata = lexer.unify(mega_content)
    lexer.change_values(mega_automata)

    with open('generated.py', 'w', encoding="utf-8") as file:
        file.write("from sara_compis1_tools.StateAFD import StateAFD\n")
        file.write("from sara_compis1_tools.lexEval import LexEval\n")
        file.write("from sara_compis1_tools.Error import Error\n")
        file.write("import sys\n\n")
        file.write("mega = [")
        for i, obj in enumerate(mega_automata):
            if obj.value:
                value_str = repr(obj.value.line) if isinstance(obj.value.line, str) else str(obj.value.line)
            else:
                value_str = None
            file.write(f"StateAFD(name='{obj.name}',transitions={obj.transitions},accepting={obj.accepting},start={obj.start}, value={value_str})")
            if i != len(mega_automata) - 1:
                file.write(",")
        file.write("]\n")

        if errors:
            file.write("errors = set([")
            for i, error in enumerate(errors):
                file.write(f"Error(line={error.line}, error='{error.error}')")
                if i != len(errors) - 1:
                    file.write(",")
            file.write("])\n\n")
        else:
            file.write("errors = set()\n\n")

        file.write("if len(sys.argv) < 2:\n\tprint('Por favor ingrese el archivo plano')\n\tsys.exit(1)\n")
        file.write("txt_file = sys.argv[1]\n\n")
        file.write("lex = LexEval(txt_file)\n")

        file.write("results = lex.evaluate(mega, errors)\n")
        file.write("lex.print_tokens(results)\n\n")

        file.write("from sara_compis1_tools.Visualizer import Visualizer\n")
        file.write("v = Visualizer()\n")
        file.write("v.draw_mega_afd(mega)\n")

